@@ -7,7 +7,7 @@ use utf8;
 use warnings;
 
 use List::Util qw(max);
-use Text::CharWidth qw(mbswidth);
+use Text::CharWidth qw(mbswidth mbwidth);
 
 require Exporter;
 our @ISA       = qw(Exporter);
@@ -16,9 +16,10 @@ our @EXPORT_OK = qw(
                        mbswidth_height
                        mbtrunc
                        mbwrap
+                       wrap
                );
 
-our $VERSION = '0.02'; # VERSION
+our $VERSION = '0.03'; # VERSION
 
 sub mbswidth_height {
     my $text = shift;
@@ -35,67 +36,135 @@ sub mbswidth_height {
     [max(@lens) // 0, $num_lines];
 }
 
-sub mbwrap {
-    my ($text, $width) = @_;
-    $width //= 80;
-
-    my @res;
-    my @p = split /(\s+)/i, $text;
-    my $col = 0;
-    my $i = 0;
-    while (my $p = shift(@p)) {
-        $i++;
-        my $num_nl = 0;
-        my $is_pb; # paragraph break
-        my $is_ws;
-        my $w;
-        #say "D:col=$col, p=$p";
-        if ($p =~ /\A\s/s) {
-            $is_ws++;
-            $num_nl++ while $p =~ s/\r?\n//;
-            if ($num_nl >= 2) {
-                $is_pb++;
-                $w = 0;
-            } else {
-                $p = " ";
-                $w = 1;
-            }
+sub _get_indent_width {
+    my ($is_mb, $indent, $tab_width) = @_;
+    my $w = 0;
+    for (split //, $indent) {
+        if ($_ eq "\t") {
+            # go to the next tab
+            $w = $tab_width * (int($w/$tab_width) + 1);
         } else {
-            $w = mbswidth($p);
-        }
-        $col += $w;
-        #say "D:col=$col, is_pb=${\($is_pb//0)}, is_ws=${\($is_ws//0)}, num_nl=$num_nl";
-
-        if ($is_pb) {
-            push @res, "\n" x $num_nl;
-            $col = 0;
-        } elsif ($col > $width+1) {
-            # remove whitespace at the end of prev line
-            if (@res && $res[-1] eq ' ') {
-                pop @res;
-            }
-
-            push @res, "\n";
-            if ($is_ws) {
-                $col = 0;
-            } else {
-                push @res, $p;
-                $col = $w;
-            }
-        } else {
-            # remove space at the end of text
-            if (@p || !$is_ws) {
-                push @res, $p;
-            } else {
-                if ($num_nl == 1) {
-                    push @res, "\n";
-                }
-            }
+            $w += $is_mb ? mbwidth($_) : 1;
         }
     }
+    $w;
+}
+
+sub _wrap {
+    my ($is_mb, $text, $width, $opts) = @_;
+    $width //= 80;
+    $opts  //= {};
+
+    # our algorithm: split into paragraphs, then process each paragraph. at the
+    # start of paragraph, determine indents (either from %opts, or deduced from
+    # text, like in Emacs) then push first-line indent. proceed to push words,
+    # while adding subsequent-line indent at the start of each line.
+
+    my $tw = $opts->{tab_width} // 8;
+    die "Please specify a positive tab width" unless $tw > 0;
+    my $optfli  = $opts->{flindent};
+    my $optfliw = _get_indent_width($is_mb, $optfli, $tw) if defined $optfli;
+    my $optsli  = $opts->{slindent};
+    my $optsliw = _get_indent_width($is_mb, $optsli, $tw) if defined $optsli;
+    my @res;
+
+    my @para = split /(\n(?:[ \t]*\n)+)/, $text;
+
+  PARA:
+    while (my ($ptext, $pbreak) = splice @para, 0, 2) {
+        my $x = 0;
+        my $y = 0;
+        my $line_has_word = 0;
+
+        # determine indents
+        my ($fli, $sli, $fliw, $sliw);
+        if (defined $optfli) {
+            $fli  = $optfli;
+            $fliw = $optfliw;
+        } else {
+            # XXX emacs can also treat ' #' as indent, e.g. when wrapping
+            # multi-line perl comment.
+            ($fli) = $ptext =~ /\A([ \t]*)\S/;
+            $fliw = _get_indent_width($is_mb, $fli, $tw);
+        }
+        if (defined $optsli) {
+            $sli  = $optsli;
+            $sliw = $optsliw;
+        } else {
+            ($sli) = $ptext =~ /\A[^\n]*\S[\n]([ \t+]*)\S/;
+            if (defined $sli) {
+                $sliw = _get_indent_width($is_mb, $sli, $tw);
+            } else {
+                $sli  = "";
+                $sliw = 0;
+            }
+        }
+        die "Subsequent indent must be less than width" if $sliw >= $width;
+
+        push @res, $fli;
+        $x += $fliw;
+
+        # process each word
+        for my $word0 ($ptext =~ /(\S+)/g) {
+            my @words;
+            my @wordsw;
+            while (1) {
+                my $wordw = $is_mb ? mbswidth($word0) : length($word0);
+                if ($wordw <= $width-$sliw) {
+                    push @words , $word0;
+                    push @wordsw, $wordw;
+                    last;
+                }
+                # truncate long word
+                if ($is_mb) {
+                    my $res = mbtrunc($text, $width-$sliw, 1);
+                    push @words , $res->[0];
+                    push @wordsw, $res->[1];
+                    $word0 = substr($word0, length($res->[0]));
+                } else {
+                    my $w2 = substr($word0, 0, $width-$sliw);
+                    push @words , $w2;
+                    push @wordsw, $width-$sliw;
+                    $word0 = substr($word0, $width-$sliw);
+                }
+            }
+
+            for my $word (@words) {
+                my $wordw = shift @wordsw;
+                if ($x + ($line_has_word ? 1:0) + $wordw <= $width) {
+                    if ($line_has_word) {
+                        push @res, " ";
+                        $x++;
+                    }
+                    push @res, $word;
+                    $x += $wordw;
+                } else {
+                    push @res, "\n", $sli, $word;
+                    $x = $sliw + $wordw;
+                    $y++;
+                }
+                $line_has_word++;
+            }
+        }
+
+        if (defined $pbreak) {
+            push @res, $pbreak;
+        } else {
+            push @res, "\n" if $ptext =~ /\n[ \t]*\z/;
+        }
+    }
+
     join "", @res;
 }
 
+sub mbwrap {
+    _wrap(1, @_);
+}
+
+sub wrap {
+    _wrap(0, @_);
+}
+
 sub mbpad {
     my ($text, $width, $which, $padchar, $is_trunc) = @_;
     if ($which) {
@@ -136,7 +205,7 @@ sub mbtrunc {
     my $end;
     while (1) {
         my $left  = substr($text, 0, $l);
-        my $right = substr($text, $l);
+        my $right = $l > length($text) ? "" : substr($text, $l);
         my $wl = mbswidth($left);
         #say "D: left=$left, right=$right, wl=$wl";
         if ($wres + $wl > $width) {
@@ -170,12 +239,12 @@ Text::WideChar::Util - Routines for text containing wide characters
 
 =head1 VERSION
 
-version 0.02
+version 0.03
 
 =head1 SYNOPSIS
 
  use Text::WideChar::Util qw(
-     mbpad mbswidth_height mbtrunc mbwrap);
+     mbpad mbswidth_height mbtrunc mbwrap wrap);
 
  # get width as well as number of lines
  say mbswidth_height("red\n红色"); # => [4, 2]
@@ -198,6 +267,8 @@ version 0.02
 This module provides routines for dealing with text containing wide characters
 (wide meaning occupying more than 1 column width in terminal).
 
+=encoding utf8
+
 =head1 FUNCTIONS
 
 =head2 mbswidth_height($text) => [INT, INT]
@@ -205,24 +276,45 @@ This module provides routines for dealing with text containing wide characters
 Like L<Text::CharWidth>'s mbswidth(), but also gives height (number of lines).
 For example, C<< mbswidth_height("foobar\nb\n") >> gives [6, 3].
 
-=head2 mbwrap($text, $width) => STR
-
-Wrap text to a specified column width.
-
-C<$width> defaults to 80 if not specified.
-
-Note: currently performance is rather abysmal (~ 1200/s on my Core i5-2400
-3.1GHz desktop for a ~ 1KB of text), so call this routine sparingly ;-).
-
-=head2 mbwrap($text, $width) => STR
+=head2 mbwrap($text, $width, \%opts) => STR
 
 Wrap C<$text> to C<$width> columns. It uses mbswidth() instead of Perl's
 length() which works on a per-character basis.
 
-Note: for text which does not have whitespaces between words, like Chinese, you
-will have to separate the words first (e.g. using L<Lingua::ZH::WordSegmenter>).
-The module also currently does not handle whitespace-like characters other than
-ASCII 32 (for example, the Chinese dot 。).
+Options:
+
+=over
+
+=item * tab_width => INT (default: 8)
+
+Set tab width.
+
+Note that tab will only have effect on the indent. Tab between text will be
+replaced with a single space.
+
+=item * flindent => STR
+
+First line indent. If unspecified, will be deduced from the first line of text.
+
+=item * slindent => STD
+
+Subsequent line indent. If unspecified, will be deduced from the second line of
+text, or if unavailable, will default to empty string (C<"">).
+
+=back
+
+Performance: ~2300/s on my Core i5-2400 3.1GHz desktop for a ~1KB of text.
+
+=head2 wrap($text, $width, \%opts) => STR
+
+Like mbwrap(), but uses character-based length() instead of column width-wise
+mbswidth(). Provided as an alternative to the venerable L<Text::Wrap>'s wrap()
+but with a different behaviour. This module's wrap() can reflow newline and its
+behavior is more akin to Emacs (try reflowing a paragraph in Emacs using
+C<M-q>).
+
+Performance: ~3100/s on my Core i5-2400 3.1GHz desktop for a ~1KB of text.
+Text::Wrap::wrap() on the other hand can go ~3800/s.
 
 =head2 mbpad($text, $width[, $which[, $padchar[, $truncate]]]) => STR
 
@@ -241,6 +333,9 @@ length(), so it can handle wide characters.
 
 Does *not* handle multiple lines.
 
+
+None are exported by default, but they are exportable.
+
 =head1 FAQ
 
 =head2 How do I truncate or pad to a certain character length (instead of column width)?
